import { Facet } from '@codemirror/state';
import { LRParser, LocalTokenGroup } from '@lezer/lr';
import { styleTags, tags } from '@lezer/highlight';
import { LRLanguage, indentNodeProp, delimitedIndent, foldNodeProp, foldInside, LanguageDescription, LanguageSupport } from '@codemirror/language';
import { autocompletion } from '@codemirror/autocomplete';
import { hoverTooltip, ViewPlugin } from '@codemirror/view';
import { setDiagnostics } from '@codemirror/lint';
import { TextDocument, RainDocument, getRainHover, CompletionItemKind, getRainCompletion, getRainDiagnostics, DiagnosticSeverity } from '@rainprotocol/rainlang';

// This file was generated by lezer-generator. You probably shouldn't edit it.
const parser = LRParser.deserialize({
  version: 14,
  states: "$UQYQPOOPwOPOOO|QPO'#CkOOQO'#Ch'#ChO!TQPO'#CgOOQO'#Cw'#CwOOQO'#Co'#CoQYQPOOP![OQO'#C]POOO)C>g)C>gO!gQPO,59VO!lQPO,59VOOQO,59R,59RO!sQPO,59ROOQO-E6m-E6mPOOO'#Cn'#CnP!zOQO,58wPOOO,58w,58wOOQO1G.q1G.qO#VQPO1G.qOOQO1G.m1G.mPOOO-E6l-E6lPOOO1G.c1G.cOOQO7+$]7+$]",
  stateData: "#`~OfOSgPQ~ORTOSTOTTOUTOVTOWTOXTO]RO`QO~OgWO~OaYO~PYOY[O~PYOh_Oi_OjaO~O^bO~OacO~PYOYdO~PYOh_Oi_OjfO~O^gO~O]`X~",
  goto: "!vlPmPPPPPPPPPpxPP!QPP!Y!`PPPPPPP!nRXP]TOQSVZ]]SOQSVZ]]ROQSVZ]Q`WRe`QVOQZQQ]SV^VZ]]UOQSVZ]",
  nodeNames: "⚠ BlockComment Program Semi Comma Assignment AngleBracketOpen Number Placeholder VariableName ParenClose Expression OpcodeOpening Opcode ParenOpen OpcodeWithOperandOpening OpcodeWithOperand AngleBracketClose",
  maxTerm: 27,
  skippedNodes: [0,1,18],
  repeatNodeCount: 2,
  tokenData: "&`~RaXY!WYZ!W]^!Wpq!Wxy!iyz!n|}!s!P!Q!x!Q!R#T!R![$`![!]%S!]!^%X!^!_%^!`!a%c!c!}%h#R#S&Z#T#o%h~!]Sf~XY!WYZ!W]^!Wpq!W~!nO^~~!sOY~~!xOS~~!{Pz{#O~#TOg~~#YQV~#U#V#`#l#m#t~#cQ!Q!R#i!R!S#i~#nQV~!Q!R#i!R!S#i~#wR!Q![$Q!c!i$Q#T#Z$Q~$VRV~!Q![$Q!c!i$Q#T#Z$Q~$eQV~!Q![$k#X#Y$t~$nQ!Q![$k#X#Y$t~$wP!Q![$z~%PPV~!Q![$z~%XOT~~%^OR~~%cOU~~%hOa~~%mUX~xy&P}!O%h!Q![%h!^!_&U!c!}%h#T#o%h~&UO]~~&ZO`~~&`OW~",
  tokenizers: [1, new LocalTokenGroup("u~RRYZ[]^az{i~aOi~~fPi~YZ[~lP!P!Qo~tOj~~", 36, 24)],
  topRules: {"Program":[0,2]},
  tokenPrec: 104
});

/**
 * @public Rainlang syntax highlighter
 */
const RainLRLanguage = LRLanguage.define({
    name: "rainlang",
    parser: parser.configure({
        props: [
            indentNodeProp.add({
                Expression: delimitedIndent({ closing: ")", align: false }),
            }),
            styleTags({
                Number: tags.number,
                Placeholder: tags.unit,
                VariableName: tags.variableName,
                Opcode: tags.keyword,
                OpcodeWithOperand: tags.keyword,
                BlockComment: tags.comment,
                AngleBracketOpen: tags.keyword,
                AngleBracketClose: tags.keyword,
                ParenOpen: tags.keyword,
                ParenClose: tags.keyword,
                Assignment: tags.separator,
                Comma: tags.punctuation,
                Semi: tags.punctuation
            }),
            foldNodeProp.add({
                Expression: foldInside,
                BlockComment(tree) { return { from: tree.from + 2, to: tree.to - 2 }; }
            })
        ]
    }),
    languageData: {
        closeBrackets: { brackets: ["(", "<"] },
        commentTokens: { block: { open: "/*", close: "*/" } }
    }
});

// const useLast = (values: readonly any[]) => values.reduce((_, v) => v, '');
// const documentUri = Facet.define<string, string>({ combine: useLast });
// const languageId = Facet.define<string, string>({ combine: useLast });
/**
 * @public A plugin that manages languages services processes.
 * Original code from [codemirror-languageserver](https://github.com/FurqanSoftware/codemirror-languageserver),
 * BSD-3-Clause License, Copyright (c) 2021, Mahmud Ridwan, All rights reserved.
 *
 * @see [LICENSE](./LICENSE) for license details.
 */
class RainLanguageServicesPlugin {
    // constructor of the class
    constructor(view, initOpMeta = "") {
        this.view = view;
        this.uri = "file:///untitled.rain";
        this.languageId = "rainlang";
        this.version = 0;
        this.textDocument = TextDocument.create(this.uri, this.languageId, this.version, this.view.state.doc.toString()),
            this.rainDocument = new RainDocument(this.textDocument, initOpMeta);
        this.processDiagnostics();
    }
    // update the instance of TexTdocument and RainDocument
    update(update) {
        if (!update.docChanged)
            return;
        else {
            TextDocument.update(this.textDocument, [{ text: this.view.state.doc.toString() }], ++this.version);
            this.rainDocument.update(this.textDocument);
            this.processDiagnostics();
        }
    }
    destroy() { }
    // update opmeta for instance of RainDocument
    async updateOpmeta(opmeta) {
        this.rainDocument.update(undefined, opmeta);
        this.processDiagnostics();
    }
    // handles calls for hover tooltip
    async handleHoverTooltip(view, position) {
        try {
            return getHoverTooltip(view.state.doc, this.rainDocument, position);
        }
        catch (err) {
            console.log(err);
            return null;
        }
    }
    // handles calls for code completion
    async handleCompletion(context, position) {
        try {
            return getCompletion(context, this.rainDocument, position);
        }
        catch (err) {
            console.log(err);
            return null;
        }
    }
    // handles calls for docuement diagnostics
    async processDiagnostics() {
        try {
            const diagnostics = await getDiagnostics(this.view.state.doc, this.rainDocument);
            this.view.dispatch(setDiagnostics(this.view.state, diagnostics));
        }
        catch (err) {
            console.log(err);
            this.view.dispatch(setDiagnostics(this.view.state, []));
        }
    }
}
/**
 * @public Provides codemirror hover tooltips.
 * Original code from [codemirror-languageserver](https://github.com/FurqanSoftware/codemirror-languageserver),
 * BSD-3-Clause License, Copyright (c) 2021, Mahmud Ridwan, All rights reserved.
 *
 * @see [LICENSE](./LICENSE) for license details.
 *
 * @param doc - Codemirror Text object
 * @param rainDocument - The RainDocument
 * @param position - Position of the textDocument to get the completion items for
 * @param setting - (optional) Language service params
 * @returns Codemirror tooltip or null
 */
function getHoverTooltip(doc, rainDocument, position, setting) {
    const result = getRainHover(rainDocument, position, setting);
    if (!result)
        return null;
    const { contents, range } = result;
    let pos = posToOffset(doc, position);
    let end;
    if (range) {
        pos = posToOffset(doc, range.start);
        end = posToOffset(doc, range.end);
    }
    if (pos === null)
        return null;
    const dom = document.createElement("div");
    dom.classList.add("documentation");
    //if (this.allowHTMLContent) dom.innerHTML = formatContents(contents);
    //else 
    dom.textContent = contents.value;
    return { pos, end, create: (_view) => ({ dom }), above: true };
}
/**
 * @public Provides codemirror code completion items.
 * Original code from [codemirror-languageserver](https://github.com/FurqanSoftware/codemirror-languageserver),
 * BSD-3-Clause License, Copyright (c) 2021, Mahmud Ridwan, All rights reserved.
 *
 * @see [LICENSE](./LICENSE) for license details.
 *
 * @param context - Completion context
 * @param rainDocument - The RainDocument
 * @param position - Position of the textDocument to get the completion items for
 * @param setting - (optional) Language service params
 * @returns Codemirror completion result or null
 */
function getCompletion(context, rainDocument, position, setting) {
    const CompletionItemKindMap = Object.fromEntries(Object.entries(CompletionItemKind).map(([key, value]) => [value, key]));
    let { pos } = context;
    // const line = state.doc.lineAt(pos);
    // let trigKind: CompletionTriggerKind = CompletionTriggerKind.Invoked;
    // let trigChar: string | undefined;
    // if (
    //     !explicit &&
    //     plugin.client.capabilities?.completionProvider?.triggerCharacters?.includes(
    //         line.text[pos - line.from - 1]
    //     )
    // ) {
    //     trigKind = CompletionTriggerKind.TriggerCharacter;
    //     trigChar = line.text[pos - line.from - 1];
    // }
    if (
    // trigKind === CompletionTriggerKind.Invoked &&
    !context.matchBefore(/\w+$/))
        return null;
    const result = getRainCompletion(rainDocument, position, setting);
    if (!result)
        return null;
    let completions = result.map(({ detail, label, kind, insertText, documentation, sortText, filterText, }) => {
        const completion = {
            label,
            detail,
            apply: (view, _comp) => {
                // re adjusting the cursor position for opcodes
                const matcher = prefixMatch([_comp]);
                const match = context.matchBefore(matcher[1]);
                if (insertText) {
                    let cursorPos = match.from + insertText.length - 1;
                    if (insertText.includes("<>"))
                        cursorPos -= 2;
                    view.dispatch({
                        changes: {
                            from: match.from,
                            to: match.to,
                            insert: insertText
                        },
                        selection: { anchor: cursorPos, head: cursorPos }
                    });
                }
                else
                    view.dispatch({
                        changes: {
                            from: match.from,
                            to: match.to,
                            insert: label
                        }
                    });
            },
            type: kind && CompletionItemKindMap[kind].toLowerCase(),
            sortText: sortText ?? label,
            filterText: filterText ?? label,
        };
        if (documentation) {
            completion.info = documentation.value;
        }
        return completion;
    });
    const [_span, _match] = prefixMatch(completions);
    const token = context.matchBefore(_match);
    if (token) {
        pos = token.from;
        const word = token.text.toLowerCase();
        if (/^\w+$/.test(word)) {
            completions = completions
                .filter(({ filterText }) => filterText.toLowerCase().startsWith(word))
                .sort(({ label: a }, { label: b }) => {
                switch (true) {
                    case a.startsWith(token.text) && !b.startsWith(token.text):
                        return -1;
                    case !a.startsWith(token.text) && b.startsWith(token.text):
                        return 1;
                }
                return 0;
            });
        }
    }
    return {
        from: pos,
        options: completions,
    };
}
/**
 * @public Provides codemirror diagnostics.
 * Original code from [codemirror-languageserver](https://github.com/FurqanSoftware/codemirror-languageserver),
 * BSD-3-Clause License, Copyright (c) 2021, Mahmud Ridwan, All rights reserved.
 *
 * @see [LICENSE](./LICENSE) for license details.
 *
 * @param doc - Codemirror Text object
 * @param rainDocument - The RainDocument
 */
async function getDiagnostics(doc, rainDocument) {
    const diagnostics = (await getRainDiagnostics(rainDocument))
        .map(({ range, message, severity }) => ({
        from: posToOffset(doc, range.start),
        to: posToOffset(doc, range.end),
        severity: {
            [DiagnosticSeverity.Error]: "error",
            [DiagnosticSeverity.Warning]: "warning",
            [DiagnosticSeverity.Information]: "info",
            [DiagnosticSeverity.Hint]: "info",
        }[severity],
        message,
    }))
        .filter(({ from, to }) => from !== null &&
        to !== null &&
        from !== undefined &&
        to !== undefined)
        .sort((a, b) => {
        switch (true) {
            case a.from < b.from:
                return -1;
            case a.from > b.from:
                return 1;
        }
        return 0;
    });
    return diagnostics;
}
/**
 * @public Converts a position to offset of a character in a codemirror text.
 * Original code from [codemirror-languageserver](https://github.com/FurqanSoftware/codemirror-languageserver),
 * BSD-3-Clause License, Copyright (c) 2021, Mahmud Ridwan, All rights reserved.
 *
 * @see [LICENSE](./LICENSE) for license details.
 *
 * @param doc - Codemirror text
 * @param pos - The position to convert
 * @returns - The offset of the provided position
 */
function posToOffset(doc, pos) {
    if (pos.line >= doc.lines)
        return;
    const offset = doc.line(pos.line + 1).from + pos.character;
    if (offset > doc.length)
        return;
    return offset;
}
/**
 * @public Converts offset of a character in a codemirror text to a position.
 * Original code from [codemirror-languageserver](https://github.com/FurqanSoftware/codemirror-languageserver),
 * BSD-3-Clause License, Copyright (c) 2021, Mahmud Ridwan, All rights reserved.
 *
 * @see [LICENSE](./LICENSE) for license details.
 *
 * @param doc - Codemirror text
 * @param offset - The offset to convert
 * @returns - The position of the provided character offset
 */
function offsetToPos(doc, offset) {
    const line = doc.lineAt(offset);
    return {
        line: line.number - 1,
        character: offset - line.from,
    };
}
/**
 * @public Constructs regexp out of completion items.
 * Original code from [codemirror-languageserver](https://github.com/FurqanSoftware/codemirror-languageserver),
 * BSD-3-Clause License, Copyright (c) 2021, Mahmud Ridwan, All rights reserved.
 *
 * @see [LICENSE](./LICENSE) for license details.
 *
 * @param completions - Codemirror completion items
 * @returns Regexp of completion items
 */
function prefixMatch(completions) {
    function toSet(chars) {
        let preamble = "";
        let flat = Array.from(chars).join("");
        const words = /\w/.test(flat);
        if (words) {
            preamble += "\\w";
            flat = flat.replace(/\w/g, "");
        }
        return `[${preamble}${flat.replace(/[^\w\s]/g, "\\$&")}]`;
    }
    const first = new Set();
    const rest = new Set();
    for (const { label } of completions) {
        const [initial, ...restStr] = label;
        first.add(initial);
        for (const char of restStr) {
            rest.add(char);
        }
    }
    const source = toSet(first) + toSet(rest) + "*$";
    return [new RegExp("^" + source), new RegExp(source)];
}
/**
 * @public Reduces an array to its last element
 * @param values - Values to reduce
 */
const useLast = (values) => values.reduce((_, v) => v, '');

/**
 * @public The facet used to store and retrive RainLanguageServicesPlugin instance from a EditorView.plugin()
 * after getting the facet value from the state associated with that EditoreView instance, so then can intract
 * with the plugin directly to update op meta for example.
 * This combination is provided for `rainlang` LanguageSupport class objectand not for `RainlangExtention` class.
 *
 * @example
 * ```typescript
 * // get value of the facet
 * const rainViewPlugin = editorView.state.facet(RainLanguageServicesFacet);
 *
 * // get the plugin instance
 * const rainPlugin = editorView.plugin(rainViewPlugin);
 *
 * // update op meta from plugin
 * rainPlugin.updateOpMeta("0x123...");
 * ```
 */
const RainLanguageServicesFacet = Facet.define({ combine: useLast });
/**
 * @public Rainlang as a codemirror extention
 * @example
 * ```typescript
 * // instantiate the extension that can be directly used as codemirror extention
 * // it is instantiated so that opmeta can be updated easily
 * const rainlangExtension = new RainlangExtention(options);
 *
 * // to update op meta
 * rainlangExtension.updateOpMeta("ox456...");
 * ```
 */
class RainlangExtension {
    /**
     * @public Constructor of the RainlangCodemirror class
     * @param config - options to for language services to include
     */
    constructor(config) {
        this.extension = [];
        this.hover = hoverTooltip((view, pos) => this.plugin?.handleHoverTooltip(view, offsetToPos(view.state.doc, pos)) ?? null);
        this.completion = autocompletion({
            override: [
                async (context) => {
                    if (this.plugin == null)
                        return null;
                    return await this.plugin?.handleCompletion(context, offsetToPos(context.state.doc, context.pos));
                },
            ],
        });
        this.extension.push(RainLRLanguage, ViewPlugin.define(view => this.plugin = new RainLanguageServicesPlugin(view, config?.initialOpMeta)));
        if (!config)
            this.extension.push(this.hover, this.completion);
        else {
            if (config?.hover)
                this.extension.push(this.hover);
            if (config?.completion)
                this.extension.push(this.completion);
        }
    }
    /**
     * @public Updates the op meta of this instance
     * @param opmeta - The new op meta
     */
    updateOpMeta(opmeta) {
        if (this.plugin)
            this.plugin.updateOpmeta(opmeta);
    }
}
/**
 * @public provides Rainlang implementation for codemirror as LanguageSupport
 * This is the standard implementation of rainlang for codemirror.
 *
 * @example
 * in order to update opmeta for it, it should be done through getting the plugin instance
 * from the EditorView with providing it the result of getting the `RainLanguageServicesFacet`
 * facet and then calling `updateOpMeta()` for it:
 * ```typescript
 * // get the `RainLanguageServicesFacet` value from editor view
 * const rainViewPlugin = view.state.facet(RainLanguageServicesFacet);
 *
 * // retrive the `RainLanguageServicesPlugin` instance
 * const rainPlugin = view.plugin(rainViewPlugin);
 *
 * // update op meta
 * rainPlugin.updateOpMeta("0x123...")
 * ```
 *
 * @param config - options to for language services to include
 * @returns rainlang LanguageSupport
 */
function rainlang(config) {
    let plugin;
    const rainViewPlugin = ViewPlugin.define(view => plugin = new RainLanguageServicesPlugin(view, config?.initialOpMeta));
    const services = [];
    const hover = hoverTooltip((view, pos) => plugin?.handleHoverTooltip(view, offsetToPos(view.state.doc, pos)) ?? null);
    const completion = autocompletion({
        override: [
            async (context) => {
                if (plugin == null)
                    return null;
                return await plugin?.handleCompletion(context, offsetToPos(context.state.doc, context.pos));
            },
        ],
    });
    if (!config)
        services.push(hover, completion);
    else {
        if (config?.hover)
            services.push(hover);
        if (config?.completion)
            services.push(completion);
    }
    return new LanguageSupport(RainLRLanguage, [
        RainLanguageServicesFacet.of(rainViewPlugin),
        rainViewPlugin,
        ...services
    ]);
}
/**
 * @public Provides LanguageDescription class object for rainlang
 */
const RainlangDescription = LanguageDescription.of({
    name: "rainlang",
    alias: ["Rain Language", "Rainlang", "RainLang", "rain"],
    extensions: [".rain", ".rainlang", ".rl"],
    load: (config) => Promise.resolve(rainlang(config)),
    support: rainlang()
});

export { RainLRLanguage, RainLanguageServicesFacet, RainLanguageServicesPlugin, RainlangDescription, RainlangExtension, getCompletion, getDiagnostics, getHoverTooltip, offsetToPos, posToOffset, prefixMatch, rainlang, useLast };
